"""Neural Network-Boosted Importance Sampling for Bayesian Statistics"""

from .sampler import Sampler
from .prior import Prior

__author__ = 'Johannes U. Lange'
__email__ = 'julange.astro@pm.me'
__version__ = '0.7.0'
